/*
 * Some JavaScript support routines for the maze
 *
 * Copyright 2010 - Michael Sinz
 */

/*
 * We like to build mazes - there are two parts to this process,
 * the first is to build the field on which we will build the maze
 * and the second is then to do the maze building.
 * We don't actually need to show the process but it looks cool so
 * we do :-)
 */
function StartMaze()
{
	var width = document.body.offsetWidth;
	var height = document.body.offsetHeight;

	var maze = new Array();

	// When we started...
	maze.startTime = new Date().getTime();

	// Zero our "stopTime" which is the amount of
	// time we spend not running the JavaScript
	maze.stopTime = 0;

	// The number of tiles per continuation (setTimeout)
	// This basically sets how many tiles between visual
	// updates of the display in the browser.
	// set this too high and you will see no updates until complete.
	// set this too low and the browser will be too slow and/or give
	// away the maze solution.
	// Set it to less than 1 and get no updates
	// (Note that IE is too slow anyway)

	// We let the URL reprogram the default step size
	// We hunt for "steps=" and then use the rest as the value
	// Yes, this is not very robust but it is just for testing
	// anyway and I have a cake to bake :-)
	maze.steps = ParseOption('steps', 11);
	
	// The initial step position
	maze.step = maze.steps;

  // Pick the random function based on the type of test
  maze.seed = ParseOption('seed', -1);
  if (maze.seed < 0)
  {
  	// Without a seed we use (default to) the random function in
    // JavaScript rather than our own...
  	maze.Random = function(range) { return Math.floor(Math.random() * range); };
  }
  else
  {
  	// Our own random number generator that we can seed...
  	// Note that this only is random for small values but then we are only
  	// using it for small values...
  	maze.Random = function(range)
  	{
  		this.seed = (this.seed * 65535 + 16383) % 8193;  		
  		return this.seed % range;
  	}
  }
	
	// The random selection of a tile coordinate (for the maze)
	maze.RandomX = function() { return this.Random((this.width - 1) / 2) * 2 + 1; };
	maze.RandomY = function() { return this.Random((this.height - 1) / 2) * 2 + 1; };

	// Where the content goes...
	maze.container = document.getElementById('content');
	maze.original = maze.container.innerHTML;

	var sample = document.getElementById('Sample');
	maze.tileWidth = sample.clientWidth;
	maze.tileHeight = sample.clientHeight;

	maze.InfoDisplay = document.getElementById('InfoDisplay');

	// The restart function (fun :-)	
	maze.Restart = function() 
	{
		this.InfoDisplay.style.display = 'none';
		this.container.innerHTML = this.original;
		setTimeout(StartMaze, 0);
	}

	maze.Ok = function()
	{
		this.InfoDisplay.style.display = 'none';
	}

	maze.ShowStats = function()
	{
		this.InfoDisplay.style.display = 'block';
	}

	maze.Hint = function()
	{
		var fromto = {fromColor:'Wall', toColor:'BlueWall'};
		ColorSolver(maze, 0, 0, fromto);
	}

	maze.UnHint = function()
	{
		var fromto = {fromColor:'BlueWall', toColor:'Wall'};
		ColorSolver(maze, 0, 0, fromto);
	}

	maze.Build = function()
	{
		FillMaze(this, this.RandomX(), this.RandomY());
	}

	// No key press until done...
	document.onkeypress = function(e) { }

	// How we do continue - via a set-timeout which then calls the given
	// function.  We use this to hide the way we "continue"
	maze.Continue = function(func) 
	{
		this.stopTime -= new Date().getTime();
		setTimeout(function()
		{
			maze.stopTime += new Date().getTime();
			func();
		}, 0);
	};

	// Compute the size of the maze
	var vMargin = 2;
	var hMargin = 2;

	var tmp = document.getElementById('top-left');
	if (tmp)
	{
		vMargin += tmp.clientHeight;
		hMargin += tmp.clientWidth;
	}

	tmp = document.getElementById('bottom-right');
	if (tmp)
	{
		vMargin += tmp.clientHeight;
		hMargin += tmp.clientWidth;
	}

	width = Math.floor((width - hMargin) / maze.tileWidth);
	height = Math.floor((height - vMargin) / maze.tileHeight);

	if (width < 9) width = 9;
	if (height < 9) height = 9;

	if (width / 2 == Math.floor(width / 2)) width--;
	if (height / 2 == Math.floor(height / 2)) height--;

	maze.width = width;
	maze.height = height;
	
	document.title = 'Simple Maze - Setup...';

	maze.Finished = function()
	{
		this.endTime = new Date().getTime();
	
		// We should now put together a nice display of what we all did
		document.maze = this;
	
		// Set the type of maze layout
		SetInfo('MazeType', this.type);
		document.title = 'Simple Maze - ' + this.type;
	
		// but I don't have that written yet...
		var deltaTime = this.endTime - this.startTime;
	
		// We call a tile just the main rooms, not the connectors
		var width = (this.width - 1) / 2;
		var height = (this.height - 1) / 2;
		var tiles = width * height;
	
		SetInfo('MazeWidth', width);
		SetInfo('MazeHeight', height);
		SetInfo('MazeSize', tiles);
	
		SetInfo('TilesPerSecond', Math.floor(tiles * 100000 / deltaTime) / 100);
	
		var insideTime = deltaTime - this.stopTime;
	
		SetInfo('InJavaScript', insideTime);
		SetInfo('InJavaScriptPerf', Math.floor(insideTime * 1000 / tiles) / 1000);
	
		SetInfo('InBrowser', this.stopTime);
		SetInfo('InBrowserPerf', Math.floor(this.stopTime * 1000 / tiles) / 1000);
	
		SetInfo('TotalTime' , deltaTime);
		SetInfo('TotalTimePerf', Math.floor(deltaTime * 1000 / tiles) / 1000);
	
		SetInfo('Steps', this.steps);
	
		this.ShowStats();
		
		// Some keyboard shortcuts - active only once the maze is complete
		document.onkeypress = function(e)
		{
			// Some browsers don't give you the event via the argument
			if (!e)
			{
				e = window.event;
			}
			if (e)
			{
				var code = e.keyCode ? e.keyCode : (e.charCode ? e.charCode : e.which);
	
				// "Trivial" ASCII lower case to upper case
				if (code > 95) code -= 32;
	
				// "O" to dismiss the statistics info
				if (code == 79) document.maze.Ok();
	
				// "H" to show the colour-based maze hinting
				if (code == 72) document.maze.Hint();
	
				// "U" to Undo the colour-based hinting
				if (code == 85) document.maze.UnHint();
	
				// "R" to rebuild the maze
				if (code == 82) document.maze.Restart();
	
				// "S" to show the statistics info
				if (code == 83) document.maze.ShowStats();
				
				// "T" for a test (flash the fill)
				if (code == 84)
				{
					document.maze.Hint();
					setTimeout(function () { setTimeout(document.maze.UnHint, 0); }, 0);
				}
			}
		}
	}

	maze.Continue(function()
		{
			SetupMaze(maze);
			
			document.title = 'Simple Maze - ' + maze.type + ' - Generating...';
			
			maze.Build();
		});
}

function ParseOption(option, defaultValue)
{
	var url = new String(document.location);
	var pos = url.lastIndexOf(option + "=");
	var result = defaultValue;
	if (pos > 0)
	{
		var t = url.substring(pos + option.length + 1);
		var p = t.indexOf('&');
		if (p > 0) t = t.substring(0, p);
		if (t >= 0) result = t;
	}
	
	return result;
}

function SetInfo(name, value)
{
	var element = document.getElementById(name);
	if (element) element.innerHTML = value;
}

function FillMaze(maze, x, y)
{
	while (true)
	{
		// If we have done enough of these operations we want to
		// use maze.Continue to continue later.
		maze.step++;
		if ((maze.step > maze.steps) && (maze.steps >= 1))
		{
			maze.step = 0;
			maze.Continue(function() {FillMaze(maze, x, y);});
			return;
		}

		maze[y][x].className = 'Floor';

		var nextPos = false;
		var tries = 0;

		while ((tries < 4) && !nextPos)
		{
			tries++;
			nextPos = PickMove(maze, x, y);
		}

		// Now we need to find a different place to continue the maze...
		//if (!nextPos) nextPos = Hunt(maze, maze.RandomX(), maze.RandomY());
		if (!nextPos) nextPos = Hunt(maze, x, y);

		// If we found nothing we must be done...
		if (!nextPos)
		{
			maze[0][1].className = 'Floor';
			maze[maze.height - 1][maze.width - 2].className = 'Floor';

			maze.Continue(function() {maze.Finished();});
			return;
		}

		// The next floor tile goes here...		
		x = nextPos.X;
		y = nextPos.Y;
	}
}

function Hunt(maze, startX, startY)
{	
	var nextPos = false;
	
	// We move in x/y and stop when we gat back to the start
	var x = startX;
	var y = startY;

	// Pick the direction of the search - there are 4 directions
	// and 2 variations of each (x-then-y or y-then-x)
	// We keep the same direction for any given search
	var direction = maze.Random(8);
	
	while (!nextPos)
	{
		switch (direction)
		{
		case 0: x += 2;
			if (x >= maze.width)
			{
				x -= maze.width - 1;
				y += 2;
				if (y >= maze.height)
				{
					y -= maze.height - 1;
				}
			}
			break;
		case 1: x -= 2;
			if (x < 0)
			{
				x += maze.width - 1;
				y += 2;
				if (y >= maze.height)
				{
					y -= maze.height - 1;
				}
			}
			break;
		case 2: x -= 2;
			if (x < 0)
			{
				x += maze.width - 1;
				y -= 2;
				if (y < 0)
				{
					y += maze.height - 1;
				}
			}
			break;
		case 3: x += 2;
			if (x >= maze.width)
			{
				x -= maze.width - 1;
				y -= 2;
				if (y < 0)
				{
					y += maze.height - 1;
				}
			}
			break;
		case 4: y += 2;
			if (y >= maze.height)
			{
				y -= maze.height - 1;
				x += 2;
				if (x >= maze.width)
				{
					x -= maze.width - 1;
				}
			}
			break;
		case 5: y -= 2;
			if (y < 0)
			{
				y += maze.height - 1;
				x += 2;
				if (x >= maze.width)
				{
					x -= maze.width - 1;
				}
			}
			break;
		case 6: y -= 2;
			if (y < 0)
			{
				y += maze.height - 1;
				x -= 2;
				if (x < 0)
				{
					x += maze.width - 1;
				}
			}
			break;
		case 7: y += 2;
			if (y >= maze.height)
			{
				y -= maze.height - 1;
				x -= 2;
				if (x < 0)
				{
					x += maze.width - 1;
				}
			}
			break;
		}

		// Try out this location...
		nextPos = FindNewSpace(maze, x, y);

		// If we are back at the start we always return
		if ((x == startX) && (y == startY))
		{
			return nextPos;
		}
	}
	
	return nextPos;
}

function FindNewSpace(maze, x, y)
{
	// This is optimized for the end case where there are few walls left
	// onto which to build floors.
	if (maze[y][x].className == 'Wall')
	{
		if ((x > 2) && (maze[y][x-2].className == 'Floor'))
		{
			return LookMove(maze, x - 2, y);
		}
		if ((x < maze.width - 3) && (maze[y][x+2].className == 'Floor'))
		{
			return LookMove(maze, x + 2, y);
		}
		if ((y > 2) && (maze[y-2][x].className == 'Floor'))
		{
			return LookMove(maze, x, y - 2);
		}
		if ((y < maze.height - 3) && (maze[y+2][x].className == 'Floor'))
		{
			return LookMove(maze, x, y + 2);
		}
	}

	return false;
}

function PickMove(maze, x, y)
{
	return CheckMove(maze, x, y, maze.Random(4));
}

// Check all four directions, picking one at random to start with.
function LookMove(maze, x, y)
{
	var d = maze.Random(4);
	var nextPos = false;
	
	for (var i = 0; (i < 4) && !nextPos; i++)
	{
		nextPos = CheckMove(maze, x, y, d);

		d++;

		while (d > 3) d -= 4;
	}

	return nextPos;
}

function CheckMove(maze, x, y, d)
{
	if ((d == 0) && (x > 2) && (maze[y][x-2].className == 'Wall'))
	{
		x--;
		maze[y][x].className = 'Floor';
		x--;
		return {X:x, Y:y};
	}
	if ((d == 1) && (x < maze.width - 3) && (maze[y][x+2].className == 'Wall'))
	{
		x++;
		maze[y][x].className = 'Floor';
		x++;
		return {X:x, Y:y};
	}
	if ((d == 2) && (y > 2) && (maze[y-2][x].className == 'Wall'))
	{
		y--;
		maze[y][x].className = 'Floor';
		y--;
		return {X:x, Y:y};
	}
	if ((d == 3) && (y < maze.height - 3) && (maze[y+2][x].className == 'Wall'))
	{
		y++;
		maze[y][x].className = 'Floor';
		y++;
		return {X:x, Y:y};
	}

	return false;
}

// A trick to showing the maze solution - colour one of the connected black
// areas as a BlueWall and thus showing the path which is along the interface
// between the Wall and BlueWall
function ColorSolver(maze, x, y, fromto)
{
	var newPos = {X:x, Y:y};

	while (newPos)
	{
		x = newPos.X;
		y = newPos.Y;
		maze[y][x].className = fromto.toColor;

		newPos = false;

		if ((x > 0) && (maze[y][x-1].className == fromto.fromColor))
		{
			newPos = ColorSolverFork(newPos, maze, x-1, y, fromto);
		}
		if ((y > 0) && (maze[y-1][x].className == fromto.fromColor))
		{
			newPos = ColorSolverFork(newPos, maze, x, y-1, fromto);
		}
		if ((x < maze.width - 1) && (maze[y][x+1].className == fromto.fromColor))
		{
			newPos = ColorSolverFork(newPos, maze, x+1, y, fromto);
		}
		if ((y < maze.height - 1) && (maze[y+1][x].className == fromto.fromColor))
		{
			newPos = ColorSolverFork(newPos, maze, x, y+1, fromto);
		}
	}
}

function ColorSolverFork(newPos, maze, x, y, fromto)
{
	if (newPos)
	{
		setTimeout(function() 
		{
			ColorSolver(maze, x, y, fromto);
		}, 0);
	}
	else
	{
		newPos = {X:x, Y:y};
	}

	return newPos;
}
